
#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

from block2 import *
from block2.su2 import *
from pyscf import lib, ao2mo
import numpy as np

class MP(lib.StreamObject):
    def __init__(self, mf, frozen=None, mo_coeff=None, mo_occ=None, mp_order=2):
        if mo_coeff is None:
            mo_coeff = mf.mo_coeff
        if mo_occ is None:
            mo_occ = mf.mo_occ
        self.mol = mf.mol
        self._scf = mf
        self.converged = False
        self.mo_coeff = mo_coeff
        self.mo_occ = mo_occ
        self.e_corr = None
        self.e_corrs = []
        self.nroots = 1
        self.frozen = frozen
        self.verbose = self.mol.verbose
        self.mp_order = mp_order
        self.stdout = self.mol.stdout

    @property
    def e_tot(self):
        return np.asarray(self.e_corr) + self._scf.e_tot
    
    def _get_random_mps(self, hamil, name, ref=False):
        if ref:
            info = CASCIMPSInfo(hamil.n_sites, hamil.vacuum, self.target, hamil.basis, 1, 0, 1)
            info.set_bond_dimension(1)
        else:
            info = MPSInfo(hamil.n_sites, hamil.vacuum, self.target, hamil.basis)
            info.set_bond_dimension(2000)
        info.tag = name
        mps = MPS(hamil.n_sites, 0, 2)
        mps.initialize(info)
        mps.random_canonicalize()
        mps.tensors[0].normalize()
        mps.save_mutable()
        info.save_mutable()
        if mps.center == 0 and mps.dot == 2:
            mps.move_left(hamil.opf.cg, None)
        mps.center = 0
        return mps
    
    def _mps_addition(self, hamil, name, mpoa, mpsa, mpob, mpsb):

        fmps = self._get_random_mps(hamil, name)

        lme = MovingEnvironment(mpoa, fmps, mpsa, "LME")
        lme.init_environments(False)
        rme = MovingEnvironment(mpob, fmps, mpsb, "RME")
        rme.init_environments(False)
        linear = Linear(None, lme, rme, VectorUBond([2000]), VectorUBond([2000]), VectorDouble([0]))
        linear.eq_type = EquationTypes.FitAddition
        linear.cutoff = 0
        linear.iprint = max(min(self.verbose - 4, 3), 0)
        linear.linear_conv_thrds = VectorDouble([1E-20])
        linear.solve(1, True, 0)

        return fmps
    
    def _build_hamiltonian(self, fcidump, ci_order, no_trans=False):

        big_left_orig = CSFBigSite(self.n_inactive, ci_order, False,
            fcidump, VectorUInt8([0] * self.n_inactive), max(min(self.verbose - 4, 3), 0))
        big_right_orig = CSFBigSite(self.n_external, ci_order, True,
            fcidump, VectorUInt8([0] * self.n_external), max(min(self.verbose - 4, 3), 0))
        big_left = SimplifiedBigSite(big_left_orig,
            NoTransposeRule(RuleQC()) if no_trans else RuleQC())
        big_right = SimplifiedBigSite(big_right_orig,
            NoTransposeRule(RuleQC()) if no_trans else RuleQC())
        hamil = HamiltonianQCBigSite(self.vacuum, self.n_orbs,
            VectorUInt8([0] * self.n_orbs), fcidump, big_left, big_right)
        self.n_sites = hamil.n_sites
        return hamil
    
    def _solve_linear(self, lmpo, rmpo, bra, ket):

        lme = MovingEnvironment(lmpo, bra, bra, "LME")
        lme.init_environments(False)
        lme.delayed_contraction = OpNamesSet.normal_ops()
        lme.cached_contraction = False
        rme = MovingEnvironment(rmpo, bra, ket, "RME")
        rme.init_environments(False)
        linear = Linear(lme, rme, None, VectorUBond([2000]), VectorUBond([2000]), VectorDouble([0]))
        linear.cutoff = 0
        linear.iprint = max(min(self.verbose - 4, 3), 0)
        linear.linear_conv_thrds = VectorDouble([1E-20])
        return linear.solve(1, True, 0)
    
    def _expectation(self, mpo, bra, ket):

        ime = MovingEnvironment(mpo, bra, ket, "IME")
        ime.init_environments(False)
        return Expect(ime, 2000, 2000).solve(False)

    def kernel(self):

        # Global
        Random.rand_seed(123456)
        scratch = './nodex'
        n_threads = lib.num_threads()
        fcidump_tol = 1E-13
        memory = lib.param.MAX_MEMORY * 1E6
        init_memory(isize=int(memory * 0.05),
            dsize=int(memory * 0.95), save_dir=scratch)
        Global.threading = Threading(
            ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global,
            n_threads, n_threads, 1)
        Global.threading.seq_type = SeqTypes.Nothing
        Global.frame.fp_codec = DoubleFPCodec(1E-16, 1024)
        Global.frame.load_buffering = False
        Global.frame.save_buffering = False
        Global.frame.use_main_stack = False
        Global.frame.minimal_disk_usage = True
        if self.verbose >= 5:
            print(Global.frame)
            print(Global.threading)

        # FCIDUMP
        h1e = self.mo_coeff.T @ self._scf.get_hcore() @ self.mo_coeff
        e_core = self.mol.energy_nuc()
        h1e = h1e.flatten()
        g2e = ao2mo.restore(8, ao2mo.kernel(self.mol, self.mo_coeff), self.mol.nao)
        h1e[np.abs(h1e) < fcidump_tol] = 0
        g2e[np.abs(g2e) < fcidump_tol] = 0
        na, nb = self.mol.nelec
        fcidump = FCIDUMP()
        fcidump.initialize_su2(self.mol.nao, na + nb, abs(na - nb), 1, e_core, h1e, g2e)
        error = fcidump.symmetrize(VectorUInt8([0] * self.mol.nao))
        if self.verbose >= 5:
            print('symm error = ', error)

        self.n_orbs = self.mol.nao
        ci_order = 2

        # Hamiltonian
        self.n_inactive = len(self.mo_occ[self.mo_occ > 1])
        self.n_external = len(self.mo_occ[self.mo_occ <= 1])
        assert self.n_inactive + self.n_external == self.n_orbs

        self.vacuum = SU2(0)
        self.target = SU2(na + nb, abs(na - nb), 0)

        hamil = self._build_hamiltonian(fcidump, ci_order)

        self.hamil = hamil

        ket0 = self._get_random_mps(hamil, "KET0", True)

        # MPO
        mpo = MPOQC(hamil, QCTypes.NC)
        mpo = SimplifiedMPO(mpo, RuleQC(), True)

        # DMRG
        me = MovingEnvironment(mpo, ket0, ket0, "DMRG")
        me.delayed_contraction = OpNamesSet.normal_ops()
        me.cached_contraction = True
        me.save_partition_info = True
        me.init_environments(False)
        dmrg = DMRG(me, VectorUBond([2000]), VectorDouble([0]))
        dmrg.davidson_conv_thrds = VectorDouble([1E-20])
        dmrg.cutoff = 0
        dmrg.iprint = max(min(self.verbose - 4, 3), 0)
        ener = dmrg.solve(1, True, 0.0)

        self.converged = True
        self.e_corr = ener - self._scf.e_tot
        self.e_corrs.append(ener - self._scf.e_tot)

        lib.logger.note(self, 'E(MP1) = %.16g  E_corr = %.16g', self.e_tot, self.e_corr)

        self.mps = ket0

        if self.mp_order == 1:
            return self.e_corr, ket0

        # FCIDUMP
        dm1 = self.make_rdm1(ket0).copy(order='C')
        fd_dyall = DyallFCIDUMP(fcidump, self.n_inactive, self.n_external)
        fd_dyall.initialize_from_1pdm_su2(dm1)
        error = fd_dyall.symmetrize(VectorUInt8([0] * self.mol.nao))
        if self.verbose >= 5:
            print('symm error = ', error)
        
        hm_dyall = self._build_hamiltonian(fd_dyall, ci_order, False)
        hamil = self._build_hamiltonian(fcidump, ci_order, True)

        # Left MPO
        lmpo = MPOQC(hm_dyall, QCTypes.NC)
        lmpo = SimplifiedMPO(lmpo, RuleQC(), True)
        lmpo.const_e -= self.e_tot
        lmpo = lmpo * -1

        # Right MPO
        rmpo = MPOQC(hamil, QCTypes.NC)
        rmpo = SimplifiedMPO(rmpo, NoTransposeRule(RuleQC()), True)
        rmpo.const_e -= self.e_tot

        # Identity MPO
        impo = IdentityMPO(hamil.basis, hamil.basis, hamil.vacuum, hamil.opf)
        impo = SimplifiedMPO(impo, Rule())

        bra = self._get_random_mps(hamil, "BRA")
        self.e_corr += self._solve_linear(lmpo, rmpo, bra, ket0)
        self.e_corrs.append(self.e_corr - self.e_corrs[-1])

        lib.logger.note(self, 'E(MP2) = %.16g  E_corr = %.16g', self.e_tot, self.e_corr)

        dp01 = self._expectation(impo, bra, ket0)
        ket1 = self._mps_addition(hamil, "KET1", impo, bra, (-dp01) * impo, ket0)
        dp11 = self._expectation(impo, ket1, ket1)

        # fac = 1 / np.sqrt(dp11 + 1), 1 / np.sqrt(dp11 + 1)
        fac = np.sqrt(1 - dp11), 1
        # mps1 = fac[0] * ket0 + fac[1] * ket1
        mps1 = self._mps_addition(hamil, "MPS1", fac[0] * impo, ket0, fac[1] * impo, ket1)

        self.mps = mps1

        if self.mp_order == 2:
            return self.e_corr, self.mps

        hex1 = self._expectation(rmpo, ket1, ket1)
        h0ex1 = -self._expectation(lmpo, ket1, ket1)

        self.e_corr += hex1 - h0ex1 - self.e_corrs[0] * dp11
        self.e_corrs.append(self.e_corr - self.e_corrs[-1])

        lib.logger.note(self, 'E(MP3) = %.16g  E_corr = %.16g', self.e_tot, self.e_corr)

        if self.mp_order == 3:
            return self.e_corr, self.mps

        ci_order = 4

        hm_dyall = self._build_hamiltonian(fd_dyall, ci_order, True)
        hamil = self._build_hamiltonian(fcidump, ci_order, True)

        # Left MPO
        lmpo = MPOQC(hm_dyall, QCTypes.NC)
        lmpo = SimplifiedMPO(lmpo, NoTransposeRule(RuleQC()), True)
        lmpo.const_e -= self.e_corrs[0] + self._scf.e_tot
        lmpo = lmpo * -1

        # Right MPO
        rmpo = MPOQC(hamil, QCTypes.NC)
        rmpo = SimplifiedMPO(rmpo, NoTransposeRule(RuleQC()), True)
        rmpo.const_e -= self._scf.e_tot

        # Identity MPO
        impo = IdentityMPO(hamil.basis, hamil.basis, hamil.vacuum, hamil.opf)
        impo = SimplifiedMPO(impo, Rule())

        bra1 = self._mps_addition(hamil, "BRA1", lmpo, ket1, rmpo, ket1)
        dp10 = self._expectation(impo, bra1, ket0)

        bra2 = self._mps_addition(hamil, "BRA2", impo, bra1, (-dp10) * impo, ket0)

        # Left MPO
        lmpo = MPOQC(hm_dyall, QCTypes.NC)
        lmpo = SimplifiedMPO(lmpo, RuleQC(), True)
        lmpo.const_e -= self._scf.e_tot
        lmpo = lmpo * -1

        bra = self._get_random_mps(hamil, "BRA")
        self.e_corr += self._solve_linear(lmpo, impo, bra, bra2) - self.e_corrs[1] * dp11
        self.e_corrs.append(self.e_corr - self.e_corrs[-1])

        lib.logger.note(self, 'E(MP4) = %.16g  E_corr = %.16g', self.e_tot, self.e_corr)

        self.mps = bra

        if self.mp_order == 4:
            return self.e_corr, self.mps
        
        dp02 = self._expectation(impo, bra, ket0)
        ket2 = self._mps_addition(hamil, "KET2", impo, bra, (-dp02) * impo, ket0)
        dp21 = self._expectation(impo, ket2, ket1)
        dp22 = self._expectation(impo, ket2, ket2)
        hex2 = self._expectation(rmpo, ket2, ket2)
        h0ex2 = -self._expectation(lmpo, ket2, ket2)

        self.e_corr += hex2 - h0ex2 - self.e_corrs[0] * dp22 - \
            2 * self.e_corrs[1] * dp21 - self.e_corrs[2] * dp11 
        self.e_corrs.append(self.e_corr - self.e_corrs[-1])

        lib.logger.note(self, 'E(MP5) = %.16g  E_corr = %.16g', self.e_tot, self.e_corr)

        self.mps = ket2

        if self.mp_order == 5:
            return self.e_corr, self.mps
    
    def make_rdm1(self, state=None, norb=None, nelec=None):
        '''
        Spin-traced one-particle density matrix in MO basis.
        dm1[p,q] = <q_alpha^\dagger p_alpha> + <q_beta^\dagger p_beta>
        '''

        if state is None:
            state = self.mps
        
        # 1PDM MPO
        pmpo = PDM1MPOQC(self.hamil, 0)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        # 1PDM
        pme = MovingEnvironment(pmpo, state, state, "1PDM")
        pme.init_environments(False)
        expect = Expect(pme, state.info.bond_dim, state.info.bond_dim)
        expect.iprint = max(min(self.verbose - 4, 3), 0)
        expect.solve(True, state.center == 0)
        dmr = expect.get_1pdm_spatial(self.n_orbs)
        dm = np.array(dmr).copy()
        dmr.deallocate()

        return dm.transpose((1, 0))

    def make_rdm2(self, state=None, norb=None, nelec=None):
        '''
        Spin-traced two-particle density matrix in MO basis
        dm2[p,q,r,s] = \sum_{sigma,tau} <p_sigma^\dagger r_tau^\dagger s_tau q_sigma>
        '''

        if state is None:
            state = self.mps
        
        # 2PDM MPO
        pmpo = PDM2MPOQC(self.hamil)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        # 2PDM
        pme = MovingEnvironment(pmpo, state, state, "2PDM")
        pme.init_environments(False)
        expect = Expect(pme, state.info.bond_dim, state.info.bond_dim)
        expect.iprint = max(min(self.verbose - 4, 3), 0)
        expect.solve(True, state.center == 0)
        dmr = expect.get_2pdm_spatial(self.n_orbs)
        dm = np.array(dmr, copy=True)

        return dm.transpose((0, 3, 1, 2))

def MP2(*args, **kwargs):
    kwargs['mp_order'] = 2
    return MP(*args, **kwargs)

def MP3(*args, **kwargs):
    kwargs['mp_order'] = 3
    return MP(*args, **kwargs)

def MP4(*args, **kwargs):
    kwargs['mp_order'] = 4
    return MP(*args, **kwargs)

def MP5(*args, **kwargs):
    kwargs['mp_order'] = 5
    return MP(*args, **kwargs)

if __name__ == '__main__':

    from pyscf import gto, scf, ci, mp

    mol = gto.M(
        atom='''
            O  0.000000  0.000000  0.000000
            H  0.758602  0.000000  0.504284
            H  0.758602  0.000000 -0.504284
        ''',
        basis='6-31g',
        verbose=3, symmetry=False, spin=0)

    mf = scf.RHF(mol).set(conv_tol=1E-12).run()
    # print(mf.scf_summary['e1'])
    # print(mf.scf_summary['e2'])
    # print(mf.mo_energy[:5].sum() * 2 + mf.mol.energy_nuc())
    mymp = MP2(mf).run()
    dm1 = mymp.make_rdm1()
    mymp = mp.MP2(mf).run()
    dm1x = mymp.make_rdm1()
    h1e = mymp.mo_coeff.T @ mymp._scf.get_hcore() @ mymp.mo_coeff
    print(np.einsum('pq,pq->', h1e, dm1))
    print(np.einsum('pq,pq->', h1e, dm1x))
    print(np.linalg.norm(dm1 - dm1x))
    quit()
    print(np.diag(dm1))
    print(np.diag(dm1x))
    from pyscf.mp.dfmp2_native import DFMP2
    mymp = DFMP2(mf).run()
    dm1yy = mymp.make_rdm1()
    dm1y = mymp.make_rdm1_unrelaxed()
    dm1z = mymp.make_rdm1_relaxed()
    print(np.linalg.norm(dm1x - dm1y))
    print(np.linalg.norm(dm1x - dm1z))
    print(np.diag(dm1y))
    print(np.diag(dm1z))
    # for xxx in dm1:
    #     print(xxx)
    # quit()
    # print(np.linalg.norm(dm1 - dm1x))
    # print(np.linalg.norm(dm1 - dm1yy))
    # print(np.linalg.norm(dm1 - dm1y))
    # print(np.diag(dm1z))
    # print(np.linalg.norm(dm1 - dm1z))



#!/usr/bin/env python3

import os
import sys
import subprocess
import platform
from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext


class CMakeExt(Extension):
    def __init__(self, name, cmdir='.'):
        Extension.__init__(self, name, [])
        self.cmake_lists_dir = os.path.abspath(cmdir)


class CMakeBuild(build_ext):
    def build_extensions(self):
        # Ensure that CMake is present and working
        try:
            out = subprocess.check_output(['cmake', '--version'])
            print(out.decode("utf-8"))
        except OSError:
            raise RuntimeError('Cannot find CMake executable!')

        print('Python3: ', sys.executable)
        print('Build Dir: ', self.build_temp)

        for ext in self.extensions:

            extdir = os.path.abspath(os.path.dirname(
                self.get_ext_fullpath(ext.name)))
            cfg = 'Release'

            cmake_args = [
                '-DCMAKE_BUILD_TYPE=%s' % cfg,
                '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(
                    cfg.upper(), extdir),
                '-DCMAKE_ARCHIVE_OUTPUT_DIRECTORY_{}={}'.format(
                    cfg.upper(), self.build_temp),
                '-DPYTHON_EXECUTABLE_HINT={}'.format(sys.executable),
                '-DUSE_MKL=ON',
                '-DMPI=ON',
                '-DBUILD_LIB=ON',
                '-DLARGE_BOND=ON',
                '-DUSE_KSYMM=OFF',
                '-DUSE_COMPLEX=ON',
                '-DUSE_SG=ON'
            ]

            # We can handle some platform-specific settings at our discretion
            if platform.system() == 'Windows':
                plat = ('x64' if platform.architecture()
                        [0] == '64bit' else 'Win32')
                cmake_args += [
                    '-DCMAKE_WINDOWS_EXPORT_ALL_SYMBOLS=TRUE',
                    '-DCMAKE_RUNTIME_OUTPUT_DIRECTORY_{}={}'.format(
                        cfg.upper(), extdir)
                ]
                if self.compiler.compiler_type == 'msvc':
                    cmake_args += [
                        '-DCMAKE_GENERATOR_PLATFORM=%s' % plat,
                    ]
                else:
                    cmake_args += [
                        '-G', 'MinGW Makefiles',
                    ]

            if not os.path.exists(self.build_temp):
                os.makedirs(self.build_temp)

            subprocess.check_call(['cmake', ext.cmake_lists_dir] + cmake_args,
                                  cwd=self.build_temp)

            subprocess.check_call(['cmake', '--build', '.', '--config', cfg, '--', '--jobs=2'],
                                  cwd=self.build_temp)


setup(name='block2-mpi',
      version='0.4.6',
      packages=find_packages(),
      ext_modules=[CMakeExt('block2')],
      cmdclass={'build_ext': CMakeBuild},
      license='LICENSE',
      description="""An efficient MPO implementation of DMRG for quantum chemistry.""",
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      author='Huanchen Zhai, Henrik R. Larsson, and Zhi-Hao Cui',
      author_email='hczhai@ucla.edu',
      url='https://github.com/block-hczhai/block2-preview',
      install_requires=[
          "mkl==2019",
          "mkl-include",
          "intel-openmp",
          "numpy",
          "cmake==3.17",
          "scipy",
          "psutil",
          "pybind11"
      ],
      scripts=["pyblock2/driver/block2main", "pyblock2/driver/gaopt",
               "pyblock2/driver/readwfn.py", "pyblock2/driver/writewfn.py"]
      )




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

from block2 import *
from block2.su2 import *
from pyscf import lib, ao2mo
import numpy as np

class CI(lib.StreamObject):
    def __init__(self, mf, frozen=None, mo_coeff=None, mo_occ=None, ci_order=2):
        if mo_coeff is None:
            mo_coeff = mf.mo_coeff
        if mo_occ is None:
            mo_occ = mf.mo_occ
        self.mol = mf.mol
        self._scf = mf
        self.converged = False
        self.mo_coeff = mo_coeff
        self.mo_occ = mo_occ
        self.e_corr = None
        self.nroots = 1
        self.frozen = frozen
        self.verbose = self.mol.verbose
        self.ci_order = ci_order
        self.stdout = self.mol.stdout
    
    @property
    def e_tot(self):
        return np.asarray(self.e_corr) + self._scf.e_tot
    
    def kernel(self, ci0=None):

        # Global
        Random.rand_seed(123456)
        scratch = './nodex'
        n_threads = lib.num_threads()
        fcidump_tol = 1E-13
        memory = lib.param.MAX_MEMORY * 1E6
        init_memory(isize=int(memory * 0.05),
            dsize=int(memory * 0.95), save_dir=scratch)
        Global.threading = Threading(
            ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global,
            n_threads, n_threads, 1)
        Global.threading.seq_type = SeqTypes.Nothing
        Global.frame.fp_codec = DoubleFPCodec(1E-16, 1024)
        Global.frame.load_buffering = False
        Global.frame.save_buffering = False
        Global.frame.use_main_stack = False
        Global.frame.minimal_disk_usage = True
        if self.verbose >= 5:
            print(Global.frame)
            print(Global.threading)
        
        # FCIDUMP
        h1e = self.mo_coeff.T @ self._scf.get_hcore() @ self.mo_coeff
        e_core = self.mol.energy_nuc()
        h1e = h1e.flatten()
        g2e = ao2mo.restore(8, ao2mo.kernel(self.mol, self.mo_coeff), self.mol.nao)
        h1e[np.abs(h1e) < fcidump_tol] = 0
        g2e[np.abs(g2e) < fcidump_tol] = 0
        na, nb = self.mol.nelec
        fcidump = FCIDUMP()
        fcidump.initialize_su2(self.mol.nao, na + nb, abs(na - nb), 1, e_core, h1e, g2e)
        error = fcidump.symmetrize(VectorUInt8([0] * self.mol.nao))
        if self.verbose >= 5:
            print('symm error = ', error)
        
        n_orbs = self.mol.nao
        self.n_orbs = self.mol.nao
        
        # Hamiltonian
        n_inactive = len(self.mo_occ[self.mo_occ > 1])
        n_external = len(self.mo_occ[self.mo_occ <= 1])
        assert n_inactive + n_external == n_orbs
        big_left = CSFBigSite(n_inactive, self.ci_order, False,
            fcidump, VectorUInt8([0] * n_inactive), max(min(self.verbose - 3, 3), 0))
        big_right = CSFBigSite(n_external, self.ci_order, True,
            fcidump, VectorUInt8([0] * n_external), max(min(self.verbose - 3, 3), 0))
        big_left = SimplifiedBigSite(big_left, RuleQC())
        big_right = SimplifiedBigSite(big_right, RuleQC())
        vacuum = SU2(0)
        target = SU2(na + nb, abs(na - nb), 0)
        hamil = HamiltonianQCBigSite(vacuum, n_orbs, VectorUInt8([0] * n_orbs), fcidump,
            big_left, big_right)
        n_sites = hamil.n_sites

        self.hamil = hamil

        # MPS
        info = MPSInfo(n_sites, vacuum, target, hamil.basis)
        info.tag = "KET%d" % self.ci_order
        info.set_bond_dimension(20000)
        mps = MPS(n_sites, 0, 2)
        mps.initialize(info)
        mps.random_canonicalize()
        mps.tensors[mps.center].normalize()
        if ci0 is not None:
            ci0.load_mutable()
            mps.tensors[mps.center].selective_copy_from(ci0.tensors[mps.center], False)
        mps.save_mutable()
        info.save_mutable()

        # MPO
        mpo = MPOQC(hamil, QCTypes.NC)
        mpo = SimplifiedMPO(mpo, RuleQC(), True)

        # DMRG
        me = MovingEnvironment(mpo, mps, mps, "DMRG")
        me.delayed_contraction = OpNamesSet.normal_ops()
        me.cached_contraction = True
        me.save_partition_info = True
        me.init_environments(False)
        dmrg = DMRG(me, VectorUBond([20000]), VectorDouble([0]))
        dmrg.davidson_conv_thrds = VectorDouble([1E-12])
        dmrg.cutoff = 0
        dmrg.iprint = max(min(self.verbose - 3, 3), 0)
        ener = dmrg.solve(1, True, 0.0)

        mps.load_mutable()
        me.contract_two_dot(0, mps)
        mps.canonical_form = "CC"
        mps.save_mutable()
        mps.save_data()

        self.converged = True
        self.e_corr = ener - self._scf.e_tot

        lib.logger.note(self, 'E(CI%s) = %.16g  E_corr = %.16g',
            "SDTQPH"[:self.ci_order], self.e_tot, self.e_corr)
        
        self.mps = mps
        
        return self.e_corr, mps
    
    def make_rdm1(self, civec=None):
        '''
        Spin-traced one-particle density matrix in MO basis.
        dm1[p,q] = <q_alpha^\dagger p_alpha> + <q_beta^\dagger p_beta>
        '''

        if civec is None:
            civec = self.mps
        
        # 1PDM MPO
        pmpo = PDM1MPOQC(self.hamil, 0)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        # 1PDM
        pme = MovingEnvironment(pmpo, civec, civec, "1PDM")
        pme.init_environments(False)
        expect = Expect(pme, civec.info.bond_dim, civec.info.bond_dim)
        expect.iprint = max(min(self.verbose - 3, 3), 0)
        expect.solve(True, civec.center == 0)
        dmr = expect.get_1pdm_spatial(self.n_orbs)
        dm = np.array(dmr).copy()
        dmr.deallocate()

        return dm.transpose((1, 0))

    def make_rdm2(self, civec=None):
        '''
        Spin-traced two-particle density matrix in MO basis
        dm2[p,q,r,s] = \sum_{sigma,tau} <p_sigma^\dagger r_tau^\dagger s_tau q_sigma>
        '''

        if civec is None:
            civec = self.mps
        
        # 2PDM MPO
        pmpo = PDM2MPOQC(self.hamil, 0)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        # 2PDM
        pme = MovingEnvironment(pmpo, civec, civec, "2PDM")
        pme.init_environments(False)
        expect = Expect(pme, civec.info.bond_dim, civec.info.bond_dim)
        expect.iprint = max(min(self.verbose - 3, 3), 0)
        expect.solve(True, civec.center == 0)
        dmr = expect.get_2pdm_spatial(self.n_orbs)
        dm = np.array(dmr, copy=True)

        return dm.transpose((0, 3, 1, 2))

def CISD(*args, **kwargs):
    kwargs['ci_order'] = 2
    return CI(*args, **kwargs)

def CISDT(*args, **kwargs):
    kwargs['ci_order'] = 3
    return CI(*args, **kwargs)

def CISDTQ(*args, **kwargs):
    kwargs['ci_order'] = 4
    return CI(*args, **kwargs)

def CISDTQP(*args, **kwargs):
    kwargs['ci_order'] = 5
    return CI(*args, **kwargs)

def CISDTQPH(*args, **kwargs):
    kwargs['ci_order'] = 6
    return CI(*args, **kwargs)

if __name__ == '__main__':

    from pyscf import gto, scf, ci

    mol = gto.M(
        atom='''
            O  0.000000  0.000000  0.000000
            H  0.758602  0.000000  0.504284
            H  0.758602  0.000000  -0.504284
        ''',
        basis='6-31g',
        verbose=3, symmetry=False, spin=0)

    mf = scf.RHF(mol).run()
    myci = CISD(mf).run()
    dm1 = myci.make_rdm1()
    myci = ci.CISD(mf).run()
    dm1x = myci.make_rdm1()
    print('1pdm diff = ', np.linalg.norm(dm1 - dm1x))
    myci = CISDT(mf).run()
    myci = CISDTQ(mf)
    _, ci0 = myci.kernel()
    myci = CISDTQP(mf)
    _, ci0 = myci.kernel(ci0=ci0)
    myci = CISDTQPH(mf)
    _, ci0 = myci.kernel(ci0=ci0)




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2021 Henrik R. Larsson <larsson@caltech.edu>
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

""" Various tools for block2

These need to be initialized first, depending on the used spin.
Example:
    > from block2 import SZ
    > import tools
    > tools.init(SZ)
    > from tools import saveMPStoDir, loadMPSfromDir, changeCanonicalForm # optional
Note: "from tools import ..." must come after the init statement

:author: Henrik R. Larsson
:author: Huanchen Zhai
"""
from typing import Union, Tuple, List
import warnings

import block2
from block2 import SZ, SU2
from block2 import FCIDUMP, TruncationTypes
from block2 import VectorDouble, VectorUInt8, VectorUBond
from block2 import TruncationTypes

import itertools
import numpy as np
import sys
import subprocess, shutil, os
try:
    from pyscf import tools, ao2mo, gto
    hasPySCF = True
except ImportError:
    hasPySCF = False

# Function definitions (set in init)
hasMPI = False
saveMPStoDir = None
loadMPSfromDir = None
changeCanonicalForm = None

def mkDir(folder: str):
    """ mkdir -p folder """
    if not os.path.exists(folder):
        try:
            os.makedirs(folder)
        except FileExistsError: # don't ask...
            pass


def init(SpinLabel: Union[SZ, SU2]):
    """ Initialize this module by setting the function definitions, depending on the used spin symmetry"""
    global hasMPI, saveMPStoDir, loadMPSfromDir, changeCanonicalForm
    if SpinLabel == SZ:
        from block2.sz import MovingEnvironment, DMRG
        from block2.sz import MPS, MPSInfo
        from block2.sz import MPO, IdentityMPO, SimplifiedMPO, RuleQC
        try:
            # parallel
            from block2.sz import MPICommunicator, ParallelMPO, ParallelRuleIdentity
        except ImportError:
            MPICommunicator = ParallelRuleIdentity = None
    else:
        assert SpinLabel == SU2
        from block2.su2 import MovingEnvironment, DMRG
        from block2.su2 import MPS, MPSInfo
        from block2.su2 import MPO, IdentityMPO, SimplifiedMPO, RuleQC
        try:
            # parallel
            from block2.su2 import MPICommunicator, ParallelMPO, ParallelRuleIdentity
        except ImportError:
            MPICommunicator = ParallelRuleIdentity = None

    # function definitions

    def saveMPStoDir(mps:MPS, mpsSaveDir:str, MPI:MPICommunicator=None):
        mps.save_data() # Important! Saves canonical form
        mkDir(mpsSaveDir)
        mps.info.save_data(f"{mpsSaveDir}/mps_info.bin")
        def copyIt(fnam:str):
            if MPI is not None and MPI.rank != 0:
                # ATTENTION: For multi node  calcs, I assume that all nodes have one global scratch dir
                return
            lastName = os.path.split(fnam)[-1]
            fst = f"cp -p {fnam} {mpsSaveDir}/{lastName}"
            # subprocess is favored but sometimes there is a problem due to memory allocation
            try:
                subprocess.call(fst.split())
            except: # May problem due to allocate memory
                print(f"# ATTENTION: saveMPStoDir with command'{fst}' failed!")
                print(f"# Error message: {sys.exc_info()[0]}")
                print(f"# Error message: {sys.exc_info()[1]}")
                print(f"# Try again with shutil")
                try:
                    # vv does not copy metadata 
                    shutil.copyfile(fnam, mpsSaveDir+"/"+lastName)
                except:
                    print(f"\t# ATTENTION: saveMPStoDir with shutil also failed")
                    print(f"\t# Error message: {sys.exc_info()[0]}")
                    print(f"\t# Error message: {sys.exc_info()[1]}")
                    print(f"\t# Try again with syscal")
                    os.system(fst)

        for iSite in range(mps.n_sites+1):
            fnam = mps.info.get_filename(False,iSite)
            copyIt(fnam)
            if MPI is not None:
                MPI.barrier()
            fnam = mps.info.get_filename(True, iSite)
            copyIt(fnam)
            if MPI is not None:
                MPI.barrier()
        for iSite in range(-1,mps.n_sites): # -1 is data
            fnam = mps.get_filename(iSite)
            copyIt(fnam)
            if MPI is not None:
                MPI.barrier()
        return

    def loadMPSfromDir(mps_info: MPSInfo,  mpsSaveDir:str, MPI:MPICommunicator=None) -> MPS:
        """  Load MPS from directory
        :param mps_info: If None, MPSInfo will be read from mpsSaveDir/mps_info.bin
        :param mpsSaveDir: Directory where the MPS has been saved
        :param MPI: MPI class (or None)
        :return: MPS state
        """
        if mps_info is None: # use mps_info.bin
            mps_info = MPSInfo(0)
            mps_info.load_data(f"{mpsSaveDir}/mps_info.bin")
        else:
            # TODO: It would be good to check if mps_info.bin is available
            #       and then compare it again mps_info input to see any mismatch
            pass 
        def copyItRev(fnam: str):
            if MPI is not None and MPI.rank != 0:
                # ATTENTION: For multi node  calcs, I assume that all nodes have one global scratch dir
                return
            lastName = os.path.split(fnam)[-1]
            fst = f"cp -p {mpsSaveDir}/{lastName} {fnam}"
            try:
                subprocess.call(fst.split())
            except: # May problem due to allocate memory, but why??? 
                print(f"# ATTENTION: loadMPSfromDir with command'{fst}' failed!")
                print(f"# Error message: {sys.exc_info()[0]}")
                print(f"# Error message: {sys.exc_info()[1]}")
                print(f"# Try again with shutil")
                try:
                    # vv does not copy metadata -.-
                    shutil.copyfile(mpsSaveDir+"/"+lastName, fnam)
                except:
                    print(f"\t# ATTENTION: loadMPSfromDir with shutil also failed")
                    print(f"\t# Error message: {sys.exc_info()[0]}")
                    print(f"\t# Error message: {sys.exc_info()[1]}")
                    print(f"\t# Try again with syscal")
                    os.system(fst)
        for iSite in range(mps_info.n_sites + 1):
            fnam = mps_info.get_filename(False, iSite)
            copyItRev(fnam)
            if MPI is not None:
                MPI.barrier()
            fnam = mps_info.get_filename(True, iSite)
            copyItRev(fnam)
            if MPI is not None:
                MPI.barrier()
        mps_info.load_mutable()
        mps = MPS(mps_info)
        for iSite in range(-1, mps_info.n_sites):  # -1 is data
            fnam = mps.get_filename(iSite)
            copyItRev(fnam)
            if MPI is not None:
                MPI.barrier()
        mps.load_data()
        mps.load_mutable()
        if MPI is not None:
            MPI.barrier()
        mps_info.bond_dim = mps.info.get_max_bond_dimension() # is not initalized
        return mps

    def changeCanonicalForm(mps: MPS,
                            parallelRule:ParallelRuleIdentity = None,
                            keepStates=None
                            ):
        """ Change the canonical form of the MPS *via* a sweep to one of the end sites of the MPS

        Useful for aligning two MPSs to completely identical canonical form.
        ATTENTION: This normalizes the MPO.

        :param mps: MPS to
        :param parallelRule: Instance of ParallelRuleIdentity, otherwise, it will not be normalized when MPI is used
        :param keepStates: Number of additional states to keep, besides maximal bond dimension
        """
        opf = block2.sz.OperatorFunctions(block2.sz.CG())
        mpo = IdentityMPO(mps.info.basis, mps.info.basis, SZ(0,0,0), opf)
        if parallelRule is not None:
            mpo = SimplifiedMPO(mpo, RuleQC(), True, False) # ParallelMPO always assumes simplification
            mpo = ParallelMPO(mpo, parallelRule)
        # vv does not work. tensors[i] != nullptr assertion
        #mps.dot = 1
        if mps.dot == 2 and mps.center == mps.n_sites-1:
            mps.center -= 1
        me = MovingEnvironment(mpo, mps, mps, "bondDimChange")
        me.init_environments(False)
        bondDim = mps.info.get_max_bond_dimension() + 100 # +100 for numerics in 2s
        dmrg = DMRG(me, VectorUBond([bondDim]), VectorDouble([0]))
        if keepStates is not None:
            assert isinstance(keepStates,int)
            dmrg.trunc_type = TruncationTypes.KeepOne * keepStates
        dmrg.cutoff = 0.0
        dmrg.quanta_cutoff = 0.0
        dmrg.forward = mps.center == 0 or mps.center == 1
        dmrg.davidson_soft_max_iter = 0
        dmrg.iprint = 0
        dmrg.solve(1, dmrg.forward, 1e-16)
        mpo.deallocate()
        return

def printDummyFunction(*args, **kwargs):
    """ Does nothing"""
    pass

def getVerbosePrinter(verbose, indent="",flush=False):
    """ Substitute of the print statement. Will be ignored if verbose is False. """

    if verbose:
        if flush:
            def _print(*args, **kwargs):
                kwargs["flush"] = True
                print(indent,*args,**kwargs)
        else:
            def _print(*args, **kwargs):
                print(indent, *args, **kwargs)
    else:
        _print = printDummyFunction
    return _print


#! /usr/bin/env python
"""
Read block2 MPS,
Write StackBlock MPS.

Needs pyblock (https://github.com/hczhai/pyblock).
Check README of block2 if you get `double free` error from python
when import pyblock and block2 in the same script.

Author:
    Huanchen Zhai
    Jun 3, 2021
"""

try:
    from pyblock2.driver.parser import parse
except ImportError:
    from parser import parse

import sys
import os
import numpy as np

if len(sys.argv) > 1:
    arg_dic = {}
    for i in range(1, len(sys.argv)):
        if sys.argv[i] in ['-su2', '-sz']:
            arg_dic[sys.argv[i][1:]] = ''
        elif sys.argv[i].startswith('-'):
            arg_dic[sys.argv[i][1:]] = sys.argv[i + 1]
        elif i == 1:
            arg_dic['config'] = sys.argv[i]
else:
    raise ValueError("""
        Usage:
            (A) python writewfn.py -config dmrg.conf -out ./out
            (B) python writewfn.py dmrg.conf
            (C) python writewfn.py -integral FCIDUMP -prefix ./scratch -su2
            (D) python writewfn.py -integral FCIDUMP -prefix ./scratch -sz
            (E) python writewfn.py ... -sym c1

        Args:
            config: StackBlock input file
            out: dir for storing StackBlock MPS (not including 'node0')

            when no config file is given/available:
                integral: path to integral file
                prefix: path to StackBlock MPS (does not include 'node0')
                su2/sz: spin-adapted or non-spin-adapted
                sym: point group name (default is d2h)
                    note that d2h also works for c1

            when the prefix is set in config file using relative path,
                it will be considered as relative to the dir of dmrg.conf
                but not relative to the current dir

        If StackBlock is run with reordered integrals, before running this script,
        a reordered FCIDUMP should be generated using
            python gaopt.py -ord RestartReorder.dat -wint FCIDUMP.NEW
        And then use python writewfn.py -integral FCIDUMP.NEW ...

        Notes:
            Only support generating onedot StackBlock Wavefunction
            with canonical form LLL....KR
            (which is the only form supported by StackBlock/OH)
    """)

scratch = '.'
integral = 'FCIDUMP'
dot = 1
su2 = True
sym = "d2h"
load_dir = "./tmp"
mps_tags = ["KET"]
if "config" in arg_dic:
    config = arg_dic["config"]
    dic = parse(config)
    dd = os.path.dirname(config)
    load_dir = dic.get("prefix", "./")
    if not os.path.isabs(load_dir):
        load_dir = ('.' if dd == '' else dd) + "/" + load_dir
    integral = dic["orbitals"]
    if not os.path.isabs(integral):
        integral = ('.' if dd == '' else dd) + "/" + integral
    su2 = "nonspinadapted" not in dic
    mps_tags = dic.get("mps_tags", "KET").split()
    sym = dic.get("sym", "d2h")
if "prefix" in arg_dic:
    load_dir = arg_dic["prefix"]
if "integral" in arg_dic:
    integral = arg_dic["integral"]
if "out" in arg_dic:
    scratch = arg_dic["out"]
if not os.path.exists(scratch + "/node0"):
    os.makedirs(scratch + "/node0")
if "su2" in arg_dic:
    su2 = True
if "sz" in arg_dic:
    su2 = False
if "sym" in arg_dic:
    sym = arg_dic["sym"]

from pyblock.qchem import DMRGDataPage, BlockHamiltonian
from pyblock.qchem.core import BlockSymmetry

from block import VectorInt, VectorMatrix, Matrix, save_rotation_matrix
from block.dmrg import SweepParams
from block.io import Global
from block.symmetry import StateInfo, VectorSpinQuantum, SpinQuantum, SpinSpace, IrrepSpace
from block.symmetry import state_tensor_product_target, state_tensor_product
from block.operator import Wavefunction


from block2 import SU2, SZ, VectorDouble, VectorUInt16
from block2 import init_memory, set_mkl_num_threads, QCTypes
from block2 import PointGroup, VectorUInt8, FCIDUMP, SeqTypes, NoiseTypes

if su2:
    SX = SU2
    from block2.su2 import MPSInfo, MPS, CG, HamiltonianQC, Expect, SparseMatrix
    from block2.su2 import StateInfo as B2StateInfo
    from block2.su2 import MPOQC, RuleQC, SimplifiedMPO, MovingEnvironment, DMRG, OperatorFunctions
else:
    SX = SZ
    from block2.sz import MPSInfo, MPS, CG, HamiltonianQC, Expect, SparseMatrix
    from block2.sz import StateInfo as B2StateInfo
    from block2.sz import MPOQC, RuleQC, SimplifiedMPO, MovingEnvironment, DMRG, OperatorFunctions

# block2 part

assert os.path.isdir(load_dir)

swap_pg = getattr(PointGroup, "swap_" + sym)
npg = {'d2h': 8, 'c2h': 4, 'c2v': 4, 'd2': 4,
       'ci': 2, 'c2': 2, 'cs': 2, 'c1': 1}[sym]
inv_swap_pg = list(np.argsort([swap_pg(x + 1) for x in range(0, npg)]) + 1)

memory = int(10 * 1E9)
init_memory(isize=int(memory * 0.1), dsize=int(memory * 0.9), save_dir=load_dir)
fcidump = FCIDUMP()
fcidump.read(integral)
n_sites = fcidump.n_sites

vaccum = SX(0)
target = SX(fcidump.n_elec, fcidump.twos, swap_pg(fcidump.isym))

orb_sym = VectorUInt8(map(swap_pg, fcidump.orb_sym))
hamil = HamiltonianQC(vaccum, n_sites, orb_sym, fcidump)
hamil.opf.seq.mode = SeqTypes.Simple


if os.path.isfile(load_dir + "/%s-mps_info.bin" % mps_tags[0]):
    print('load MPSInfo from', load_dir + "/%s-mps_info.bin" % mps_tags[0])
    mps_info = MPSInfo(0)
    mps_info.load_data(load_dir + "/%s-mps_info.bin" % mps_tags[0])
else:
    print('create MPSInfo')
    mps_info = MPSInfo(n_sites, vaccum, target, hamil.basis)
    mps_info.tag = mps_tags[0]

mps = MPS(mps_info).deep_copy(mps_info.tag + "-TMP")
mps.info.load_mutable()

max_bdim = max([x.n_states_total for x in mps.info.left_dims])
if mps.info.bond_dim < max_bdim:
    mps.info.bond_dim = max_bdim
max_bdim = max([x.n_states_total for x in mps.info.right_dims])
if mps.info.bond_dim < max_bdim:
    mps.info.bond_dim = max_bdim

mps.load_data()

cg = CG(200)
cg.initialize()
orig_cf = mps.canonical_form
if mps.canonical_form[0] == 'C' and mps.canonical_form[1] == 'R':
    mps.canonical_form = 'K' + mps.canonical_form[1:]
    mps.center = 0
elif mps.canonical_form[-1] == 'C' and mps.canonical_form[-2] == 'L':
    mps.canonical_form = mps.canonical_form[:-1] + 'S'
    mps.center = mps.n_sites - 1
while mps.center > mps.n_sites - 2:
    mps.move_left(cg, None)
while mps.center < mps.n_sites - 2:
    mps.move_right(cg, None)
if mps.canonical_form[mps.center] == 'S':
    mps.flip_fused_form(mps.center, cg, None)
print(orig_cf, '->', mps.canonical_form, mps.center)

center = mps.center

mps.save_data()
mps.load_mutable()

# pyblock part

page = DMRGDataPage(save_dir=scratch)

opts = dict(fcidump=integral, pg=sym,
            su2=su2, output_level=-1, memory=25000,
            omp_threads=1, mkl_threads=1, page=page)

hamil_cxt = BlockHamiltonian.get(**opts)
xhamil = hamil_cxt.__enter__()

site_sts = xhamil.site_state_info
swpp = SweepParams()
Global.dmrginp.load_prefix = scratch + "/node0"
Global.dmrginp.save_prefix = scratch + "/node0"

forward = True

assert center == hamil.n_sites - dot - 1

stls = []
for i in range(0, center + 1):
    qs = VectorSpinQuantum()
    ns = VectorInt()
    sti = mps.info.left_dims[i + 1]
    dx = list(range(sti.n))
    dx.sort(key=lambda j:
        (sti.quanta[j].n, sti.quanta[j].twos, inv_swap_pg[sti.quanta[j].pg]))
    for j in dx:
        qx = sti.quanta[j]
        qq = SpinQuantum(qx.n, SpinSpace(qx.twos), IrrepSpace(inv_swap_pg[qx.pg] - 1))
        qs.append(qq)
        ns.append(sti.n_states[j])
    st = StateInfo(qs, ns)
    if su2:
        st.save(forward, VectorInt(range(i + 1)), 0)
        st.save(forward, VectorInt(range(i + 1)), -1)
    else:
        st.save(forward, VectorInt(range((i + 1) * 2)), 0)
        st.save(forward, VectorInt(range((i + 1) * 2)), -1)
    stls.append(st)


strs = []
for i in range(center + dot, hamil.n_sites):
    qs = VectorSpinQuantum()
    ns = VectorInt()
    sti = mps.info.right_dims[i]
    dx = list(range(sti.n))
    dx.sort(key=lambda j:
        (sti.quanta[j].n, sti.quanta[j].twos, inv_swap_pg[sti.quanta[j].pg]))
    for j in dx:
        qx = sti.quanta[j]
        qq = SpinQuantum(qx.n, SpinSpace(qx.twos), IrrepSpace(inv_swap_pg[qx.pg] - 1))
        qs.append(qq)
        ns.append(sti.n_states[j])
    st = StateInfo(qs, ns)
    if su2:
        st.save(not forward, VectorInt(range(i, hamil.n_sites)), 0)
        st.save(not forward, VectorInt(range(i, hamil.n_sites)), -1)
    else:
        st.save(not forward, VectorInt(range(i * 2, hamil.n_sites * 2)), 0)
        st.save(not forward, VectorInt(range(i * 2, hamil.n_sites * 2)), -1)
    strs.append(st)


def swap_order_left(idx):
    dd = {}
    mps.info.load_left_dims(idx)
    mps.info.load_left_dims(idx + 1)
    l, m, r = mps.info.left_dims[idx], hamil.basis[idx], mps.info.left_dims[idx + 1]
    clm = B2StateInfo.get_connection_info(l, m, r)
    for ik in range(r.n):
        bbed = clm.n if ik == r.n - 1 else clm.n_states[ik + 1]
        dx = []
        g = 0
        for bb in range(clm.n_states[ik], bbed):
            ibba = clm.quanta[bb].data >> 16
            ibbb = clm.quanta[bb].data & 0xFFFF
            nx = l.n_states[ibba] * m.n_states[ibbb]
            dx.append((l.quanta[ibba], m.quanta[ibbb], g, g + nx))
            g += nx
        dx.sort(key=lambda x: (
            x[0].n, x[0].twos, inv_swap_pg[x[0].pg], x[1].n, x[1].twos, inv_swap_pg[x[1].pg]))
        pp = []
        for xx in dx:
            pp += list(range(xx[2], xx[3]))
        dd[r.quanta[ik]] = np.argsort(pp)
    return dd


def swap_order_right(idx):
    dd = {}
    mps.info.load_right_dims(idx)
    mps.info.load_right_dims(idx + 1)
    l, m, r = mps.info.right_dims[idx], hamil.basis[idx], mps.info.right_dims[idx + 1]
    clm = B2StateInfo.get_connection_info(m, r, l)
    for ik in range(l.n):
        bbed = clm.n if ik == l.n - 1 else clm.n_states[ik + 1]
        dx = []
        g = 0
        for bb in range(clm.n_states[ik], bbed):
            ibba = clm.quanta[bb].data >> 16
            ibbb = clm.quanta[bb].data & 0xFFFF
            nx = m.n_states[ibba] * r.n_states[ibbb]
            dx.append((m.quanta[ibba], r.quanta[ibbb], g, g + nx))
            g += nx
        dx.sort(key=lambda x: (
            x[1].n, x[1].twos, inv_swap_pg[x[1].pg], x[0].n, x[0].twos, inv_swap_pg[x[0].pg]))
        pp = []
        for xx in dx:
            pp += list(range(xx[2], xx[3]))
        dd[l.quanta[ik]] = np.argsort(pp)
    return dd


for i in range(1, center):
    stl = stls[i - 1]
    str = site_sts[i][0]
    st = state_tensor_product(stl, str)
    st.collect_quanta()
    rot = [None] * len(st.quanta)
    swod = swap_order_left(i)
    for k in range(len(st.quanta)):
        xq = st.quanta[k]
        q = SX(xq.n, xq.s.irrep, swap_pg(xq.symm.irrep + 1))
        iq = mps.tensors[i].info.find_state(q)
        if iq == -1 or iq >= mps.tensors[i].info.n:
            rot[k] = []
            continue
        mat = mps.tensors[i][iq]
        rot[k] = np.zeros((mat.m, mat.n))
        rot[k][swod[q], :] = np.array(mat, copy=True)
    mrot = [None] * len(st.quanta)
    for ir, r in enumerate(rot):
        if len(r) == 0:
            mrot[ir] = Matrix()
        else:
            mrot[ir] = Matrix(np.ascontiguousarray(r))
    mrot = VectorMatrix(mrot)
    if su2:
        save_rotation_matrix(VectorInt(range(i + 1)), mrot, 0)
        save_rotation_matrix(VectorInt(range(i + 1)), mrot, -1)
    else:
        save_rotation_matrix(VectorInt(range(2 * (i + 1))), mrot, 0)
        save_rotation_matrix(VectorInt(range(2 * (i + 1))), mrot, -1)

if su2:
    wave_sites = VectorInt(range(0, center + 1))
else:
    wave_sites = VectorInt(range(0, 2 * (center + 1)))
wll = state_tensor_product(stls[center - 1], xhamil.site_state_info[center][0])
wll.collect_quanta()
wrr = strs[0]
llrr = state_tensor_product_target(wll, wrr)
wave = Wavefunction()
target_state_info = BlockSymmetry.to_state_info([(xhamil.target, 1)])
wave.initialize(target_state_info.quanta, wll, wrr, dot == 1)
wave.clear()

assert wave.onedot == (dot == 1)

wfn = mps.tensors[center]

swl = swap_order_left(center)
swr = swap_order_right(center + 1)
for (il, ir), mat in wave.non_zero_blocks:
    xql = wll.quanta[il]
    xqr = wrr.quanta[ir]
    ql = SX(xql.n, xql.s.irrep, swap_pg(xql.symm.irrep + 1))
    qr = SX(xqr.n, xqr.s.irrep, swap_pg(xqr.symm.irrep + 1))
    q = wfn.info.delta_quantum.combine(ql, -qr)
    iq = wfn.info.find_state(q)
    if iq == -1:
        continue
    xmat = wfn[iq]
    assert xmat.m == mat.ref.shape[0]
    assert xmat.n == mat.ref.shape[1]
    xx = np.array(mat.ref, copy=True)
    xy = np.array(mat.ref, copy=True)
    xx[swl[ql], :] = np.array(xmat, copy=True)
    xy[:, swr[qr]] = np.ascontiguousarray(xx)
    f = -1 if ql.twos == -2 or ql.twos == 2 else 1
    mat.ref[:, :] = f * np.ascontiguousarray(xy)


wave.save_wavefunction_info(llrr, wave_sites, 0)
wave.save_wavefunction_info(llrr, wave_sites, -1)

swpp.save_state(not forward, 0)
